import toCloneable from './toCloneable';
/**
 * A fetch receiver
 * - For received requests, message to service worker.
 * - For messaged requests, send to the receiver. <- Intercept here.
 */
export default class Channel {
    constructor(workerContainer) {
        Object.defineProperty(this, "port", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "fulfillList", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "beActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
        Object.defineProperty(this, "registerController", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (workerContainer) => {
                const { controller } = workerContainer;
                if (!controller)
                    return;
                const { port1, port2 } = new MessageChannel();
                port1.onmessage = this.onMessage;
                this.port = port1;
                this.fulfillList = [];
                controller.postMessage({ onfetch: true }, [port2]);
                if (this.beActive)
                    port1.postMessage({ status: 'on' });
            }
        });
        /**
         * For received requests, message to the service worker.
         */
        Object.defineProperty(this, "fetch", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (...args) => {
                const { port, fulfillList } = this;
                if (!port || !fulfillList) {
                    throw new Error('Service worker not ready yet');
                }
                port.postMessage({
                    request: await toCloneable(new Request(...args)),
                    index: fulfillList.length,
                });
                return new Promise((resolve, reject) => {
                    fulfillList.push([resolve, reject]);
                });
            }
        });
        Object.defineProperty(this, "onMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (event) => {
                if (!event.data)
                    return;
                if ('request' in event.data) {
                    // eslint-disable-next-line no-void
                    void this.onRequestMessage(event);
                }
                if ('response' in event.data) {
                    this.onResponseMessage(event);
                }
                if ('status' in event.data) {
                    this.onStatusMessage(event);
                }
            }
        });
        /**
         * For messaged requests, send to the receiver.
         */
        Object.defineProperty(this, "onRequestMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: async (event) => {
                const { data: { request, index } } = event;
                const { port } = this;
                if (!port) {
                    throw new Error('Service worker not ready yet');
                }
                const response = await this.fetch(request.url, request).catch((err) => err);
                port.postMessage({
                    response: response instanceof Error ? response : await toCloneable(response),
                    index,
                });
            }
        });
        /**
         * For messaged responses, treat as a previous received request's response.
         */
        Object.defineProperty(this, "onResponseMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (event) => {
                const { data: { response, index } } = event;
                const { fulfillList } = this;
                if (!fulfillList) {
                    throw new Error('Service worker not ready yet');
                }
                const fulfill = fulfillList[index];
                if (!fulfill) {
                    throw new Error('Response came for unknown request');
                }
                if (response instanceof Error) {
                    fulfill[1](response);
                }
                else {
                    fulfill[0](new Response(response.body, response));
                }
                fulfillList[index] = null;
            }
        });
        // Resolve when the service worker responds.
        Object.defineProperty(this, "statusResolve", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "onStatusMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (event) => {
                const { data: { status } } = event;
                const { port } = this;
                if (!port) {
                    throw new Error('Service worker not ready yet');
                }
                this.beActive = status === 'on';
                const { statusResolve } = this;
                if (statusResolve) {
                    statusResolve();
                    this.statusResolve = null;
                }
                else {
                    port.postMessage({ status });
                }
            }
        });
        Object.defineProperty(this, "isActive", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => this.beActive
        });
        /**
         * Start receiving worker messages.
         */
        Object.defineProperty(this, "activate", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => this.switchToStatus('on')
        });
        /**
         * Stop receiving worker messages.
         */
        Object.defineProperty(this, "restore", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: () => this.switchToStatus('off')
        });
        this.registerController(workerContainer);
        workerContainer.addEventListener('controllerchange', () => {
            this.registerController(workerContainer);
        });
    }
    async switchToStatus(status) {
        const { port } = this;
        if (!port)
            return;
        await new Promise((resolve) => {
            this.statusResolve = resolve;
            port.postMessage({ status });
        });
    }
}
